/* staffettacomplicato_ConTurno.c */

#ifndef _THREAD_SAFE
  #define _THREAD_SAFE
#endif
#ifndef _POSIX_C_SOURCE
  #define _POSIX_C_SOURCE 200112L
#endif

#include "printerror.h"

#include <unistd.h>   /* exit() etc */
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>     /* per strerror_r  and  memset */
#include <sys/mman.h>  /* shm_* stuff, and mmap() */
#include <sys/types.h>
#include <signal.h>
#include <sys/time.h> /* timeval{} for select() */
#include <time.h> /* timespec{} for pselect() */
#include <limits.h> /* for OPEN_MAX */
#include <errno.h>
#include <sys/wait.h>
#include <assert.h>
#include <pthread.h>
#include <semaphore.h>
#include <stdint.h>
#include <inttypes.h>
#include <fcntl.h>
#include <sys/stat.h>

#include <pthread.h> 

#include "DBGpthread.h"


#define NUM_ALFA 5
#define NUM_BETA 4
#define TURNOALFA 0
#define TURNOBETA 1

typedef struct sharedBuffer {
		pthread_mutex_t mutex;
		pthread_cond_t cond_alfa;
		pthread_cond_t cond_beta;

		/*per indicare il turno tra i corridori alfa*/
		int turno_tra_alfa;
		/*pre indicare il turno tra i corridori beta*/
		int turno_tra_beta;
		int turno_gruppo;
} SharedBuffer;
SharedBuffer *P;


void alfa(int index){
	char Alabel[128]; 
	char Alabelsignal[128]; 

	sprintf( Alabel, "A%d", index);
	sprintf( Alabelsignal, "A%d->B", index);

	while(1){
		DBGpthread_mutex_lock(&P->mutex, Alabel);
		
		while(P->turno_tra_alfa!=index || P->turno_gruppo!=TURNOALFA){
			DBGpthread_cond_wait(&P->cond_alfa,&P->mutex, Alabel);
		}
		
		printf("alfa %d parte\n",index);
		DBGpthread_mutex_unlock(&P->mutex, Alabel);
		sleep(1);
		DBGpthread_mutex_lock(&P->mutex, Alabel);
		printf("alfa %d ha completato il giro\n",index);
		
		/* controllo se tutti hanno fatto un giro, se e' cosi' azzero il contatore */
		/* preparo il successivo corridore alfa */
		if(P->turno_tra_alfa == NUM_ALFA-1)
			P->turno_tra_alfa=0;
		else
			P->turno_tra_alfa++;
		
		/* il prossimo corridore deve essere un beta */
		P->turno_gruppo=TURNOBETA;
		
		DBGpthread_cond_broadcast(&P->cond_beta, Alabelsignal);
		DBGpthread_mutex_unlock(&P->mutex, Alabel);
	}
	pthread_exit(NULL); /* e' l'unico pthread del processo */ 
}


void beta(int index){
	char Blabel[128]; 
	char Blabelsignal[128]; 

	sprintf( Blabel, "B%d", index);
	sprintf( Blabelsignal, "B%d->A", index);

	while(1){
		DBGpthread_mutex_lock(&P->mutex, Blabel);
		
		while(P->turno_tra_beta!=index || P->turno_gruppo!=TURNOBETA){
			DBGpthread_cond_wait(&P->cond_beta,&P->mutex, Blabel);
		}
		
		printf("beta %d parte\n",index);
		DBGpthread_mutex_unlock(&P->mutex, Blabel);
		sleep(1);
		DBGpthread_mutex_lock(&P->mutex, Blabel);
		printf("beta %d ha completato il giro\n",index);
		
		/* controllo se tutti hanno fatto un giro, se e' cosi' azzero il contatore */
		/* preparo il successivo corridore beta */
		if(P->turno_tra_beta == NUM_BETA-1)
			P->turno_tra_beta=0;
		else
			P->turno_tra_beta++;
		
		/* il prossimo corridore deve essere un alfa */
		P->turno_gruppo=TURNOALFA;
		
		DBGpthread_cond_broadcast(&P->cond_alfa, Blabelsignal);
		DBGpthread_mutex_unlock(&P->mutex, Blabel);
	}
	pthread_exit(NULL); /* e' l'unico pthread del processo */ 
}

int main(){

	int shmfd, rc, i;
	pid_t pid;
	int shared_seg_size = sizeof(SharedBuffer);
	pthread_mutexattr_t mattr;
	pthread_condattr_t cvattr;

	shmfd = shm_open( "/pedala", O_CREAT /*| O_EXCL*/ | O_RDWR, S_IRWXU );
	if (shmfd < 0) {
		perror("In shm_open()");
		exit(1);
	}
	/* adjusting mapped file size (make room for the whole segment to map) */
	rc = ftruncate(shmfd, shared_seg_size);
	if (rc != 0) {
		perror("ftruncate() failed");
		exit(1);
	}

	P = (SharedBuffer*)mmap(NULL, shared_seg_size,
		PROT_READ|PROT_WRITE, MAP_SHARED, shmfd, 0);
	if( P==MAP_FAILED )
	 		PrintERROR_andExit(errno,"mmap  failed");

	rc=pthread_mutexattr_init(&mattr);
	if( rc ) PrintERROR_andExit(rc,"pthread_mutexattr_init  failed");
	rc=pthread_mutexattr_setpshared(&mattr,PTHREAD_PROCESS_SHARED);
	if( rc ) PrintERROR_andExit(rc,"pthread_mutexattr_setpshared  failed");
	rc=pthread_condattr_init(&cvattr);
	if( rc ) PrintERROR_andExit(rc,"pthread_condattr_init  failed");
	rc=pthread_condattr_setpshared(&cvattr,PTHREAD_PROCESS_SHARED);
	if( rc ) PrintERROR_andExit(rc,"pthread_condattr_setpshared  failed");

	rc = pthread_cond_init(&P->cond_alfa, &cvattr );
	if( rc ) PrintERROR_andExit(rc,"pthread_cond_init failed");
	rc = pthread_cond_init(&P->cond_beta, &cvattr );
	if( rc ) PrintERROR_andExit(rc,"pthread_cond_init failed");
	rc = pthread_mutex_init(&P->mutex, &mattr ); 
	if( rc ) PrintERROR_andExit(rc,"pthread_mutex_init failed");
	
	
	/*per indicare il turno dei corridori alfa*/
	P->turno_tra_alfa=0;
	/*pre indicare il turno dei corridori beta*/
	P->turno_tra_beta=0;
	P->turno_gruppo = TURNOALFA;

	/*creo i thread alfa*/
	for(i=0;i<NUM_ALFA;i++) {
		pid=fork();
		if(pid<0) 
			PrintERROR_andExit(errno,"fork failed");
		else if(pid==0) {
			alfa(i);
			exit(0);
		}
	}


	/*creo i thread beta*/
	for(i=0;i<NUM_BETA;i++) {
		pid=fork();
		if(pid<0) 
			PrintERROR_andExit(errno,"fork failed");
		else if(pid==0) {
			beta(i);
			exit(0);
		}
	}

	return 0;
}


